create or replace function public.return_data_model(end_node text)
returns table(
	source_schema	varchar
	, source_object	varchar
	, source_object_type	varchar
	, source_object_stage	INT
	, dependent_schema	varchar
	, dependent_object	varchar
	, dependent_object_type	varchar
	, dependent_object_stage	INT
	, dependent_object_create_code TEXT
)

as $function$
BEGIN
	RETURN QUERY
	
	-- stage 1: determine all dependencies for the input parameter end_node
		WITH RECURSIVE dependencies(dependent_schema, dependent_object, dependent_object_type, source_schema, source_object, source_object_type) AS (
	    select DISTINCT
		dependent_ns.nspname::varchar as dependent_schema
		, dependent_object.relname::varchar as dependent_object 
		, case
			when dependent_object.relkind = 'r' then 'table'
			when dependent_object.relkind = 'v' then 'view'
			when dependent_object.relkind = 'm' then 'materialized view'
			end::varchar as dependent_object_type
		, source_ns.nspname::varchar as source_schema
		, source_object.relname::varchar as source_object
		, case
			when source_object.relkind = 'r' then 'table'
			when source_object.relkind = 'v' then 'view'
			when source_object.relkind = 'm' then 'materialized view'
			end::varchar as source_object_type
		FROM
			pg_depend 
		JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid 
		JOIN pg_class as dependent_object ON pg_rewrite.ev_class = dependent_object.oid 
		JOIN pg_class as source_object ON pg_depend.refobjid = source_object.oid 
		JOIN pg_attribute ON pg_depend.refobjid = pg_attribute.attrelid 
		    AND pg_depend.refobjsubid = pg_attribute.attnum 
		JOIN pg_namespace dependent_ns ON dependent_ns.oid = dependent_object.relnamespace
		JOIN pg_namespace source_ns ON source_ns.oid = source_object.relnamespace
		WHERE 
			source_ns.nspname = 'public'
			AND
			dependent_object.relname ~ end_node
			AND
			pg_attribute.attnum > 0
	
	UNION ALL
	 
	 	SELECT
		distinct dependent_ns.nspname::varchar as dependent_schema
		, dependent_object.relname::varchar as dependent_object 
		, case
			when dependent_object.relkind = 'r' then 'table'
			when dependent_object.relkind = 'v' then 'view'
			when dependent_object.relkind = 'm' then 'materialized view'
			end::varchar as dependent_object_type
		, source_ns.nspname::varchar as source_schema
		, source_object.relname::varchar as source_object
		, case
			when source_object.relkind = 'r' then 'table'
			when source_object.relkind = 'v' then 'view'
			when source_object.relkind = 'm' then 'materialized view'
			end::varchar as source_object_type
		FROM
			dependencies, pg_depend 
		JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid 
		JOIN pg_class as dependent_object ON pg_rewrite.ev_class = dependent_object.oid 
		JOIN pg_class as source_object ON pg_depend.refobjid = source_object.oid 
		JOIN pg_attribute ON pg_depend.refobjid = pg_attribute.attrelid 
		    AND pg_depend.refobjsubid = pg_attribute.attnum 
		JOIN pg_namespace dependent_ns ON dependent_ns.oid = dependent_object.relnamespace
		JOIN pg_namespace source_ns ON source_ns.oid = source_object.relnamespace
		WHERE 
			source_ns.nspname = 'public'
			AND
			dependent_object.relname = dependencies.source_object
			AND
			pg_attribute.attnum > 0
	),
	
	-- stage 2: determine all unique dependencies (stage 1 creates some duplicates)
	agg_dependencies as(
		SELECT DISTINCT
			dependencies.dependent_schema
			, dependencies.dependent_object
			, dependencies.dependent_object_type
			, dependencies.source_schema
			, dependencies.source_object
			, dependencies.source_object_type
	FROM
		dependencies),
	
	-- stage 3: determine the processing stage of processing of each element in the data model
	-- explanation: elements with no dependencies (i.e. tables) are set to stage 1, all depending views are processed in stages 2 to n
	object_stages(source_object, dependent_object, stage) as (
	select distinct
		a.source_object
		, a.dependent_object
		, 1 as stage
		
	from
		agg_dependencies as a
	
	left join agg_dependencies as b
	on a.source_object = b.dependent_object
	
	group by
		a.source_object
		, a.dependent_object
	
		having count(b.dependent_object) = 0
	
	union all
		select
			a.dependent_object
			, a.source_object
			, object_stages.stage + 1 as stage
		from
			object_stages, agg_dependencies as a
		where
			a.source_object = object_stages.source_object),
	
	-- stage 4: consolidate data model element processing stages
	stage_information as (
	select distinct
		object_stages.source_object
		, max(stage) as stage
	from
		object_stages
	group by
		object_stages.source_object)

-- ##################################################
-- MAIN QUERY
-- ##################################################
	
	SELECT DISTINCT
		agg_dependencies.source_schema
		, agg_dependencies.source_object
		, agg_dependencies.source_object_type
		, source_stage.stage as source_object_stage
		, agg_dependencies.dependent_schema
		, agg_dependencies.dependent_object
		, agg_dependencies.dependent_object_type
		, dependent_stage.stage as dependent_object_stage
		, pg_get_viewdef(agg_dependencies.dependent_object, true) as dependent_object_create_code
	FROM
		agg_dependencies
	left join stage_information as source_stage
	on agg_dependencies.source_object = source_stage.source_object
	left join stage_information as dependent_stage
	on agg_dependencies.dependent_object = dependent_stage.source_object
	order by
		source_stage.stage asc;
END;



$function$ language plpgsql